use stick::Event;
use tokio::{io::AsyncWriteExt, net::TcpStream};

#[derive(Default, Debug)]
pub struct GamepadRepresentation {
    // 1 byte
    action_a: bool,
    action_b: bool,
    action_x: bool,
    action_y: bool,
    arrow_up: bool,
    arrow_left: bool,
    arrow_right: bool,
    arrow_down: bool,
    // 2 byte
    bumper_left: bool,
    bumper_right: bool,
    menu_right: bool,
    menu: bool,
    menu_left: bool,
    cam: bool,
    joy: bool,
    // 3 bytes
    cam_x: i8,
    // 4 bytes
    cam_y: i8,
    // 5 bytes
    joy_x: i8,
    // 6 bytes
    joy_y: i8,
    // 7 bytes
    trigger_left: u8,
    // 8 bytes
    trigger_right: u8,
    need_refresh: bool,
}

impl GamepadRepresentation {
    pub fn process_event(&mut self, event: Event) {
        self.need_refresh = true;
        match event {
            Event::ActionB(action_a) => self.action_a = action_a, // This is inversed on XBOX controller
            Event::ActionA(action_b) => self.action_b = action_b,
            Event::ActionH(action_x) => self.action_x = action_x,
            Event::ActionV(action_y) => self.action_y = action_y,
            Event::Up(up) => self.arrow_up = up,
            Event::Left(left) => self.arrow_left = left,
            Event::Right(right) => self.arrow_right = right,
            Event::Down(down) => self.arrow_down = down,
            Event::BumperL(bumper_left) => self.bumper_left = bumper_left,
            Event::BumperR(bumper_right) => self.bumper_right = bumper_right,
            Event::MenuL(menu_left) => self.menu_left = menu_left,
            Event::Exit(menu) => self.menu = menu,
            Event::MenuR(menu_right) => self.menu_right = menu_right,
            Event::Cam(cam) => self.cam = cam,
            Event::Joy(joy) => self.joy = joy,
            Event::JoyX(joy_x) => self.joy_x = (joy_x * i8::MAX as f64) as i8,
            Event::JoyY(joy_y) => self.joy_y = (-joy_y * i8::MAX as f64) as i8,
            Event::CamX(cam_x) => self.cam_x = (cam_x * i8::MAX as f64) as i8,
            Event::CamY(cam_y) => self.cam_y = (-cam_y * i8::MAX as f64) as i8,
            Event::TriggerL(trigger_l) => self.trigger_left = (trigger_l * u8::MAX as f64) as u8,
            Event::TriggerR(trigger_r) => self.trigger_right = (trigger_r * u8::MAX as f64) as u8,
            _ => return,
        }
    }
    pub async fn send(&mut self, socket: &mut TcpStream) {
        if self.need_refresh {
            self.need_refresh = false;
            let bytes: [u8; 8] = self.into();
            if let Err(err) = socket.write_all(&bytes).await {
                println!("{}", err);
            };
        }
    }
}

fn u8_from_bools(bools: &[bool]) -> u8 {
    let mut result = 0;
    for (i, &bool) in bools.iter().enumerate() {
        result |= (bool as u8) << i;
    }
    result
}

impl From<&mut GamepadRepresentation> for [u8; 8] {
    fn from(g: &mut GamepadRepresentation) -> Self {
        let mut bytes = [0; 8];
        bytes[0] = u8_from_bools(&[
            g.action_a,
            g.action_b,
            g.action_x,
            g.action_y,
            g.arrow_down,
            g.arrow_left,
            g.arrow_right,
            g.arrow_up,
        ]);
        bytes[1] = u8_from_bools(&[
            g.bumper_left,
            g.bumper_right,
            g.menu_left,
            g.menu,
            g.menu_right,
            g.joy,
            g.cam,
        ]);
        bytes[2..3].copy_from_slice(&g.joy_x.to_le_bytes());
        bytes[3..4].copy_from_slice(&g.joy_y.to_le_bytes());
        bytes[4..5].copy_from_slice(&g.cam_x.to_le_bytes());
        bytes[5..6].copy_from_slice(&g.cam_y.to_le_bytes());
        bytes[6..7].copy_from_slice(&g.trigger_left.to_le_bytes());
        bytes[7..8].copy_from_slice(&g.trigger_right.to_le_bytes());
        bytes
    }
}
