use crate::prelude::*;
use core::sync::atomic::{AtomicBool, AtomicI8, AtomicU8};
use embassy_net::{tcp::TcpSocket, IpAddress, Ipv4Address};
use embedded_svc::ipv4::{Ipv4Addr, SocketAddr, SocketAddrV4};
use smoltcp::wire::IpEndpoint;

pub static GAMEPAD: GamepadRepresentation = GamepadRepresentation::default();

const ADDRESS: IpEndpoint =
    IpEndpoint::new(IpAddress::Ipv4(Ipv4Address::new(192, 168, 106, 27)), 1069);

#[derive(Debug)]
pub struct GamepadRepresentation {
    // 1 byte
    action_a: AtomicBool,
    action_b: AtomicBool,
    action_x: AtomicBool,
    action_y: AtomicBool,
    arrow_up: AtomicBool,
    arrow_left: AtomicBool,
    arrow_right: AtomicBool,
    arrow_down: AtomicBool,
    // 2 byte
    bumper_left: AtomicBool,
    bumper_right: AtomicBool,
    menu_right: AtomicBool,
    menu: AtomicBool,
    menu_left: AtomicBool,
    cam: AtomicBool,
    joy: AtomicBool,
    // 3 bytes
    cam_x: AtomicI8,
    // 4 bytes
    cam_y: AtomicI8,
    // 5 bytes
    joy_x: AtomicI8,
    // 6 bytes
    joy_y: AtomicI8,
    // 7 bytes
    trigger_left: AtomicU8,
    // 8 bytes
    trigger_right: AtomicU8,
}

fn bools_from_u8(results: &[&AtomicBool], mut value: u8) {
    for result in results {
        result.store((value & 1) == 1, Ordering::Relaxed);
        value >>= 1;
    }
}

impl GamepadRepresentation {
    // Cannot be done with trait derivation because Rust trait system not yet supporting const function
    // This will probably change soon with RFC 2632.

    const fn default() -> Self {
        GamepadRepresentation {
            action_a: AtomicBool::new(false),
            action_b: AtomicBool::new(false),
            action_x: AtomicBool::new(false),
            action_y: AtomicBool::new(false),
            arrow_up: AtomicBool::new(false),
            arrow_left: AtomicBool::new(false),
            arrow_right: AtomicBool::new(false),
            arrow_down: AtomicBool::new(false),
            bumper_left: AtomicBool::new(false),
            bumper_right: AtomicBool::new(false),
            menu_right: AtomicBool::new(false),
            menu: AtomicBool::new(false),
            menu_left: AtomicBool::new(false),
            cam: AtomicBool::new(false),
            joy: AtomicBool::new(false),
            cam_x: AtomicI8::new(0),
            cam_y: AtomicI8::new(0),
            joy_x: AtomicI8::new(0),
            joy_y: AtomicI8::new(0),
            trigger_left: AtomicU8::new(0),
            trigger_right: AtomicU8::new(0),
        }
    }

    pub fn action_a(&self) -> bool {
        self.action_a.load(Ordering::Relaxed)
    }

    pub fn action_b(&self) -> bool {
        self.action_b.load(Ordering::Relaxed)
    }

    pub fn action_x(&self) -> bool {
        self.action_x.load(Ordering::Relaxed)
    }

    pub fn action_y(&self) -> bool {
        self.action_y.load(Ordering::Relaxed)
    }

    pub fn arrow_up(&self) -> bool {
        self.arrow_up.load(Ordering::Relaxed)
    }

    pub fn arrow_right(&self) -> bool {
        self.arrow_right.load(Ordering::Relaxed)
    }

    pub fn arrow_left(&self) -> bool {
        self.arrow_left.load(Ordering::Relaxed)
    }

    pub fn arrow_down(&self) -> bool {
        self.arrow_down.load(Ordering::Relaxed)
    }

    pub fn bumper_left(&self) -> bool {
        self.bumper_left.load(Ordering::Relaxed)
    }

    pub fn bumper_right(&self) -> bool {
        self.bumper_right.load(Ordering::Relaxed)
    }

    pub fn menu_right(&self) -> bool {
        self.menu_right.load(Ordering::Relaxed)
    }

    pub fn menu(&self) -> bool {
        self.menu.load(Ordering::Relaxed)
    }

    pub fn menu_left(&self) -> bool {
        self.menu_left.load(Ordering::Relaxed)
    }

    pub fn cam_x(&self) -> i8 {
        self.cam_x.load(Ordering::Relaxed)
    }

    pub fn cam_y(&self) -> i8 {
        self.cam_y.load(Ordering::Relaxed)
    }

    pub fn joy_x(&self) -> i8 {
        self.joy_x.load(Ordering::Relaxed)
    }

    pub fn joy_y(&self) -> i8 {
        self.joy_y.load(Ordering::Relaxed)
    }

    pub fn trigger_left(&self) -> u8 {
        self.trigger_left.load(Ordering::Relaxed)
    }

    pub fn trigger_right(&self) -> u8 {
        self.trigger_right.load(Ordering::Relaxed)
    }

    fn update(&self, bytes: &[u8; 8]) {
        bools_from_u8(
            &mut [
                &self.action_a,
                &self.action_b,
                &self.action_x,
                &self.action_y,
                &self.arrow_down,
                &self.arrow_left,
                &self.arrow_right,
                &self.arrow_up,
            ],
            bytes[0],
        );
        bools_from_u8(
            &mut [
                &self.bumper_left,
                &self.bumper_right,
                &self.menu_left,
                &self.menu,
                &self.menu_right,
                &self.joy,
                &self.cam,
            ],
            bytes[1],
        );

        self.joy_x
            .store(i8::from_le_bytes([bytes[2]]), Ordering::Relaxed);
        self.joy_y
            .store(i8::from_le_bytes([bytes[3]]), Ordering::Relaxed);
        self.cam_x
            .store(i8::from_le_bytes([bytes[4]]), Ordering::Relaxed);
        self.cam_y
            .store(i8::from_le_bytes([bytes[5]]), Ordering::Relaxed);
        self.trigger_left
            .store(u8::from_le_bytes([bytes[6]]), Ordering::Relaxed);
        self.trigger_right
            .store(u8::from_le_bytes([bytes[7]]), Ordering::Relaxed);
    }
}

#[embassy_executor::task]
pub async fn gamepad_task(stack: &'static Stack<WifiDevice<'static>>) {
    let mut rx_buffer = [0; 256];
    let mut tx_buffer = [0; 256];
    let mut buf = [0; 8];

    loop {
        if stack.is_link_up() {
            break;
        }
        Timer::after(Duration::from_millis(500)).await;
    }

    println!("Waiting to get IP address...");
    loop {
        if let Some(config) = stack.config_v4() {
            println!("Got IP: {}", config.address);
            break;
        }
        Timer::after(Duration::from_millis(500)).await;
    }

    loop {
        let mut socket = TcpSocket::new(stack, &mut rx_buffer, &mut tx_buffer);
        if let Err(e) = socket.connect(ADDRESS).await {
            println!("{:?}", e);
            Timer::after(Duration::from_millis(10)).await;
            continue;
        }
        println!("Connected to {}", ADDRESS);
        loop {
            let Ok(_) = socket.read(&mut buf).await else {
                break;
            };
            GAMEPAD.update(&buf);
        }
        socket.close()
    }
}
